
#include <Arduino.h>
/*
 * Control Software for 3T Cable Bot 
 *  by Florian Bauer, 24.3.2025
 *  
 * Hardware:
 *  Mechanics:
 *   Cube Scaffold Structure 
 *   3 Cable pulleys on top: 
 *      Back-Left: Motor 4
 *      Back-Right: Motor 2
 *      Frond-Middle: Motor 1
 *   1 Cable pulley on the middle bottom: Motor 3
 *   4 Motors with Quadrature Encoder
 *   4 cable drums
 *  
 *  Electronics:  
 *    Arduino Mega
 *    Adafruit L293D Motor Driver Shield für 4 Motoren 
 *    2 x IC-MD to read out Enocder Motors
 *   
 * Setup:
 *    
 * Description:
 *      With this Sketch the position of an triangular shaped Endeffector haning in a Cube structure can be controlled
 *      by spooling cables       
 *      
 *      The cable length is measured as Motor position. 
 *      Encoder Motors are used and controlled via PID controllers
 *      Encoder readout is done with IC-MD counter chips
 *      
 * Anschlüsse:     
 * ic-MD 1  SCK(15) MISO(13) MOSI(14) CS(16)
 * MEGA      52         50     51       17
 * ic-MD 2  SCK(15) MISO(13) MOSI(14) CS(16)
 * MEGA      52         50     51       16
 *  
 * Encoder-Eingänge:
 *  Konfiguration  2 x TTL wie in Table 20 im Data-Sheet:
 *  ic-MD 1          AP(4) AN(5)    BN(7) CP(8)  alle Pull-Up 4.7k
 *  Motorencoder      A1     B1      A2  B2
 *  ic-MD 2          AP(4) AN(5)    BN(7) CP(8)  alle Pull-Up 4.7k
 *  Motorencoder      A3     B3      A4  B4   
 *  
* Motor-Mapping zur Seilkbot3T-Library
* Motor 4 -> l1 Links hinten oben - Endeffektor
* Motor 2 -> l2  Rechts hinten oben  - Endeffektor
* Motor 1 -> l3  Vorne mitte Vorderkante oben - Endeffektor
* Motor 3 -> l4  Unten Mitte Grundfläche- Endeffektor
* 
* Nullposition:
*  x0 238
*  y0 238
*  z0 300
*  
* Commands:
*   Z<x0>,<y0>,<z0>
*   m<x>,<y>,<z> - Position anfahren (x,y,z in cm)
*   - <mot>      - 500 counts aufwickeln
*   + <mot>      - 500 counts abwickeln
*   t <mot>      - teste motor 
*   i            - position ausgeben
*   ?            - position und weitere daten ausgeben
*   M<mot><dpos> - Motor um dpos verstellen
*   f<factor>    - setze CPR-Multiplier (Counts per revolution)
*   z            - Stop, Zählerstand als aktuelle Position setzen
*   S            - wie z
*   c            - Kreisen (Baustelle)   
*   
*  Anleitung:
*    Motor-Nummern und Wickelrichtungen prüfen
*    Manuell oder mit + und - Befehlen Endeffektor in die Mitte setzen und Seile straffen
*    Arduino Resetten (Konsole neu starten)
*    oder Position ausmessen und mit Z eingeben
*    
*  To Do:
*     - Forward Kinematic 
*     - Seilspannungsmessung
*     
*/

#define XMAX 550
#define YMAX 550
#define ZMAX 550
 
#include <SPI.h>
#include <SIKTEC_SPI.h>

#include <AFMotor.h>
#include <PID_v1_bc.h>

//#define CNTCFG TWO_TTL_24BIT
#define CNTCFG TWO_TTL_16BIT

#include "Seilbot3Tlib.h"

//#define OVERFLOW_HANDLER
#include "ICMD.h"

//double kp = 6, ki = .0150, kd = 0.04; //0.01;             // modify for optimal performance
double kp = 12, ki = .00, kd = 0.0; //0.01;  
double input1 = 0, output1 = 0, setpoint1 = 0;
PID myPID1(&input1, &output1, &setpoint1, kp, ki, kd, DIRECT);  
double input2 = 0, output2 = 0, setpoint2 = 0;
PID myPID2(&input2, &output2, &setpoint2, kp, ki, kd, DIRECT);  
double input3 = 0, output3 = 0, setpoint3 = 0;
PID myPID3(&input3, &output3, &setpoint3, kp, ki, kd, DIRECT);  
double input4 = 0, output4 = 0, setpoint4 = 0;
PID myPID4(&input4, &output4, &setpoint4, kp, ki, kd, DIRECT);  

AF_DCMotor mot1(1,MOTOR12_2KHZ);
AF_DCMotor mot2(2,MOTOR12_2KHZ);
AF_DCMotor mot3(3,MOTOR12_2KHZ);
AF_DCMotor mot4(4,MOTOR12_2KHZ);

int l1_0,l2_0,l3_0,l4_0;

//245
//Seilbot    seilbot(474.0,50.0, 500.0, 1.0, 7.8, 235);
   Seilbot    seilbot(475.0,475.0, 30.0,10.0, 1.0, 7.8, 235);
  //  Seilbot    seilbot(430.0,40.0, 480.0, 1.0, 7.8, 235);  
void pwmOut(int i,int out) {     
                
  if (out > 0) {                         // if REV > encoderValue motor move in forward direction.    
     switch(i)
     {
      case 0: mot1.run(FORWARD); mot1.setSpeed(out); break;
      case 1: mot2.run(FORWARD); mot2.setSpeed(out); break;
      case 2: mot3.run(FORWARD); mot3.setSpeed(out); break;
      case 3: mot4.run(FORWARD); mot4.setSpeed(out); break;

     //   case 2: mot3.run(FORWARD); mot3.setSpeed(out); break;
     }
  }
  else {
     switch(i) {
     case 0 : mot1.run(BACKWARD); mot1.setSpeed(abs(out)); break;
     case 1: mot2.run(BACKWARD); mot2.setSpeed(abs(out)); break;
     case 2 : mot3.run(BACKWARD); mot3.setSpeed(abs(out)); break;
     case 3: mot4.run(BACKWARD); mot4.setSpeed(abs(out)); break;
     }

  }
}


//#define SOFTWARE_COUNTER
//#define OVERFLOW_HANDLER - geht nich nicht richtig

#define COUNTER_ENABLE 16
#define COUNTER_DONE 15


//GREEN IC-MD-Pin 15
#define SLAVE_SCK_PIN 52 
//VIOLET IC-MD-Pin 13
#define SLAVE_MISO_PIN 50 
// BLUE IC-MD-Pint 14
#define SLAVE_MOSI_PIN 51 

//YELLOW The spi device SC pin:
//#define SLAVE_CS_PIN (10-4)
// IC-MD-Pin 16
#define SLAVE_CS_PIN 17
#define SLAVE_CS_PIN2 16

volatile ICMD  * icmd;
volatile ICMD  * icmd2;
int runmode=1;

void setup (void)
{
  Serial.begin(250000); //initialize serial comunication 
  Serial.println("setup");
  byte buf [3];

  icmd = new ICMD( SLAVE_CS_PIN, SLAVE_SCK_PIN,SLAVE_MISO_PIN,SLAVE_MOSI_PIN); //,20); //53);
  icmd->readManufacturerID();
  icmd2 = new ICMD( SLAVE_CS_PIN2, SLAVE_SCK_PIN,SLAVE_MISO_PIN,SLAVE_MOSI_PIN); //,20); //53);
  icmd2->readManufacturerID();   

  setpoint1=0;
  setpoint2=0;
  setpoint3=0;
  setpoint4=0;
  
  pinMode(COUNTER_DONE,INPUT);
  pinMode(COUNTER_ENABLE, OUTPUT);
  digitalWrite(COUNTER_ENABLE,LOW);
  pinMode(SLAVE_CS_PIN2,OUTPUT);
  digitalWrite(SLAVE_CS_PIN2,HIGH);

  pinMode(32,INPUT);
  
  myPID1.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID1.SetSampleTime(1);  // refresh rate of PID controller
  myPID1.SetOutputLimits(-255, 255); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.
  
  myPID2.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID2.SetSampleTime(1);  // refresh rate of PID controller
  myPID2.SetOutputLimits(-255, 255); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.
  
  myPID3.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID3.SetSampleTime(1);  // refresh rate of PID controller
  myPID3.SetOutputLimits(-255, 255); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.
  
  myPID4.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID4.SetSampleTime(1);  // refresh rate of PID controller
  myPID4.SetOutputLimits(-255, 255); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.

  seilbot.setZeroPosition(238.0,238.0,300.0);
  seilbot.calcDeltaCounts(238.0,238.0,300.0,&l1_0,&l2_0,&l3_0,&l4_0);
   
  Serial.println("setup done");
} // end of setup

volatile byte response_old[10];

String readString;

int step= -1;
long told=0;

void loop (void)
{  

  long  rawCount;
  long  count;
  int changed=0;
  byte  buf[128];
  int   ref[3];
  byte  c='a';
  char * ptr;
  
  uint32_t cnt=0;
  //byte response[10];
  byte * response,response2;

 
 
  // lese daten:  #
  response = icmd->readData();
  response = icmd2->readData();
     //icmd->printCounter();
     //Serial.println("");
     //icmd->readManufacturerID();

  while (Serial.available()) { //Check if the serial data is available.
    delay(3);                  // a small delay
    char c = Serial.read();  // storing input data
    readString += c;         // accumulate each of the characters in readString
  }
 
   if (readString.length() >0) { //Verify that the variable contains information  
      readString.toCharArray(buf, 100);
      if (strncmp(buf,"M",1)==0)
      {
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           int imot=atoi(ptr);
           ptr = strtok(NULL,",");
           if( ptr != NULL)
           {
            int pos=atof(ptr);
            Serial.print("Set Motor "); Serial.print(imot); Serial.print(" to "); Serial.println(pos);
            switch(imot)
            {
              case 1: setpoint1 +=pos; break;
              case 2: setpoint2 +=pos; break;
              case 3: setpoint3 +=pos; break;
              case 4: setpoint4 +=pos; break;
            }
           }  
         }  
      }
      if (strncmp(buf,"t",1)==0)
      {
          ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           int imot=atoi(ptr);
           ptr = strtok(NULL,",");
           if( ptr != NULL)
           {
             int out=atoi(ptr);
             Serial.print("Test Motor "); Serial.print(imot);
             Serial.print( ": ");
             Serial.println(out);
             pwmOut(imot-1,out);
             delay(1000);
             pwmOut(imot-1,0);
           }
         } 
         
      }
      if (strncmp(buf,"i",1)==0)
      {
        Serial.println(input1);
         Serial.println(input2); Serial.println(input3); Serial.println(input4);
      }
      if (strncmp(buf,"?",1)==0)
      {
        double l1,l2,l3,l4;
        Serial.print(seilbot.x); Serial.print(" ");
         Serial.print(seilbot.y); Serial.print(" ");
          Serial.print(seilbot.z); Serial.print(" ");
          seilbot.calcLengths(seilbot.x,seilbot.y,seilbot.z,&l1,&l2,&l3,&l4);
                  Serial.print(l1); Serial.print(" ");
         Serial.print(l4); Serial.print(" ");
          Serial.print(l2); Serial.print(" ");
          Serial.print(l3); Serial.println(" ");
      }
       if (strncmp(buf,"f",1)==0)
      {
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           double factor=atof(ptr);
           if (factor<0.1) factor=0.1;
           seilbot.setFactor(factor);
          
    
         }  
      }
      if (strncmp(buf,"z",1)==0)
      {
        setpoint1=input1;
        setpoint2=input2;
        setpoint3=input3;
        setpoint4=input4;
        //icmd->resetCounter(0);
        //icmd->resetCounter(1);
        //icmd2->resetCounter(0);
        //icmd2->resetCounter(1);
      }
       if (strncmp(buf,"Z",1)==0)
      {
        double x0,y0,z0;
        x0=atof(ptr);
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           y0=atof(ptr);
           ptr = strtok(NULL,",");
           if( ptr != NULL)
           {
            z0=atof(ptr);
            seilbot.setZeroPosition(x0,y0,z0);
            seilbot.calcDeltaCounts(x0,y0,z0,&l1_0,&l2_0,&l3_0,&l4_0);
             setpoint1=0;
             setpoint2=0;
             setpoint3=0;
             setpoint4=0;
             icmd->resetCounter(0);
             icmd->resetCounter(1);
             icmd2->resetCounter(0);
             icmd2->resetCounter(1);
        
           }  
         }  
      }
      if (strncmp(buf,"S",1)==0) // Stop!
      {
           setpoint1=input1;
           setpoint2=input2;
           setpoint3=input3;
           setpoint4=input4;        
      }
      if (strncmp(buf,"+",1)==0)
      {
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           int mot =atoi(ptr);
           switch(mot)
           {
            case 1:setpoint1 += 500; break;
            case 2:setpoint2 += 500;break;
            case 3:setpoint3 += 500; break;
            case 4: setpoint4 += 500; break;
           }          
         }
      }    
      if (strncmp(buf,"-",1)==0)
      {
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           int mot =atoi(ptr);
            switch(mot)
           {
            case 1:setpoint1 -= 500; break;
            case 2:setpoint2 -= 500; break;
            case 3:setpoint3 -= 500; break;
            case 4: setpoint4 -= 500; break;
           }          
         }
      }    
      if (strncmp(buf,"c",1)==0)
      {
        step=0;
      }
      if (strncmp(buf,"s",1)==0)
      {
        runmode=1-runmode;
      }
      if (strncmp(buf,"m",1)==0 )
      {
        
        double x,y,z;
        int pos1,pos2,pos3,pos4;
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           x=atof(ptr);
           ptr = strtok(NULL,",");
           if( ptr != NULL)
           {
            y=atof(ptr);
            ptr = strtok(NULL,",");
            if( ptr != NULL)
            {
               z=atof(ptr);
               if (x>=0 && x<XMAX && y>=0 && y<YMAX && z>=0 && z<ZMAX)
               {
                  seilbot.calcDeltaCounts(x,y,z,&pos4,&pos2,&pos1,&pos3);
                  seilbot.setPos(x,y,z);
                  Serial.print(" x="); Serial.print(x);
                  Serial.print(" y="); Serial.print(y);
                  Serial.print(" z="); Serial.println(z);
                  Serial.print(" POS1="); Serial.print(pos1);
                  Serial.print(" POS2="); Serial.print(pos2);
                  Serial.print(" POS3="); Serial.print(pos3);
                  Serial.print(" POS4="); Serial.println(pos4);
                  setpoint1 = pos1- l1_0;
                  setpoint2 = pos2 - l2_0;
                  setpoint3 = pos3 - l3_0;
                  setpoint4 = pos4 - l4_0;
               }
               else
                  Serial.println("Error: Out of Range");
            }
           }  
         }  
      }
      if (strncmp(buf,"d",1)==0 )
      {
        int d;
         ptr = strtok(buf+1, ",");
         if(ptr != NULL)
         {
           d=atoi(ptr);

           setpoint1 -= d;
           setpoint2 += d;
           setpoint3 -= d;
         }
      }
      changed=1;
   }
       
   if(step >=0)
   { int x,y,z; int pos1,pos2,pos3,pos4;
   double dphi= 2.0 * M_PI / 15;
    if (step==0) 
    {
      told = millis();
      step=1;
    }
    else
    if(millis()-told>=400)
    { 
     told=millis();  
    
    x = (int) (238+150*cos( (double)step * dphi));
    y = (int) (100+150*sin ( (double)step * dphi));
    Serial.print(x); Serial.print(", "); Serial.println(y);
    z=238;
                              
    seilbot.calcDeltaCounts(x,y,z,&pos4,&pos2,&pos1,&pos3);
    seilbot.setPos(x,y,z);
                 
       
    setpoint1 = pos1;
    setpoint2 = pos2;
    setpoint3 = pos3;
    setpoint4 = pos4;
                 
    changed=1;
    step ++;
    if(step==31) step= -1;
    }
   }
    
    
  // Änderung auswerten:
  if(changed || icmd->hasChanged() || icmd2->hasChanged())
  {

   // input1 = (int)(0.8*input + 0.2*icmd->getCounter(response,1));
    

    //setpoint = map(pulseIn(32,HIGH),0,2000,-20000,20000);
   
    input1 = icmd->getCounter(0);
    input2 = icmd->getCounter(1);
    input3 = icmd2->getCounter(0);
    input4 = icmd2->getCounter(1);

    Serial.print(" M1 ");  Serial.print(setpoint1); Serial.print("/");  Serial.print(input1);
    Serial.print(" M2 ");  Serial.print(setpoint2); Serial.print("/");  Serial.print(input2);
    Serial.print(" M3 ");  Serial.print(setpoint3); Serial.print("/");  Serial.print(input3);
    Serial.print(" M4 ");  Serial.print(setpoint4); Serial.print("/");  Serial.println(input4);

    myPID1.Compute();                 // calculate new output
   //    output1  *= vel1;
   //   output1 /= velmax;
    if(runmode)pwmOut(0,output1);

  
    
    myPID2.Compute();                 // calculate new output
   //   output2  *= vel2;
   //   output2 /= velmax;
    if(runmode) pwmOut(1,output2);
    
    myPID3.Compute();                 // calculate new output
   //      output3  *= vel3;
   //   output3 /= velmax;
    if(runmode)pwmOut(2,output3);
    
    myPID4.Compute();                 // calculate new output
    //      output4  *= vel4;
    //  output4 /= velmax;
    if(runmode) pwmOut(3,output4);    
   /* 
    Serial.print("A "); icmd->printCounter();
    Serial.print("B "); icmd2->printCounter();
    Serial.print(input1);
    Serial.print(" ");
    Serial.print(setpoint1);
    Serial.print(" ");
    Serial.print(output1);
    Serial.print(" ");
    Serial.print(icmd->turns[0]);
    Serial.print(" ");
    Serial.print(icmd->turns[1]);
    Serial.println();
  */
    count= icmd->getCounter(response,1);

  }
  readString="";
} // end of loop
