/*
 * Backward Kinematics für Seilbot mit 4 Seilen an 4 Ecken
 *
 * Hinterer Rahmen:
 *   A (Breite)
 * o--------o
 * |        |
 * z        |  B (Höhe)
 * |        |
 * o-- x -- o
 * 
 * Linker Rahmen; (Blick von links)
 *   A (Breite)
 * o--------o
 * |        |
 * z        |  B (Höhe)
 * |        |
 * o-- y -- o
 * 
 * l1: Links hinten oben - Endeffektor
 * l2: Rechts hinten oben  - Endeffektor
 * l3: Vorne mitte Vorderkante oben - Endeffektor
 * l4: Unten Mitte Grundfläche- Endeffektor
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "seilbot3Tlib.h"



Seilbot::Seilbot(double A,double B, double c, double d, double delta, double r, int cpr)
{
    double h; // (Dreiecks)Hoehe Endeffektor
    double m2,h2;
    this->x0=0.0; this->x=0.0;
    this->y0=0.0; this->y=0.0;
    this->z0=0.0; this->z=0.0;
    
    this->A=A;
    this->B=B;
    this->c=c;
    this->d=d;
    m2=c*c/ ( 2.0 - 2*cos(120.0/180*M_PI));
    h2=0.75*c*c;
    h=sqrt(h2);
    this->m=sqrt(m2);
    this->g = h - this->m;
    
    this->delta=delta;
    this->r=r;
    // !!!!!!!!!!!!!!!
    this->cpr=cpr;
    this->factor=1.0;
    this->lpr= (this->r + this->delta/2)*M_PI;
    this->l1_0 = this->A / 2;
    this->calcLengths(A/2,A/2,0,&this->l1_0,&this->l2_0,&this->l3_0,&this->l4_0);    
}

void Seilbot::setFactor(double f)
{
  this->factor=f;
}

void Seilbot::setPos(double x, double y, double z)
{
   this->x=x; this->y=y; this->z=z;
}
void Seilbot::calcLengths(double x, double y, double z, double * l1,double * l2, double *l3, double * l4)
{
   double t1  = (this->B - z); t1 = t1*t1;
   double t2  = (y-this->g); t2 = t2*t2;
   double t3  = (x-this->c / 2); t3 = t3*t3;
   *l1 =sqrt(t1+t2+t3);
   double t4 = (this->A-x-this->c/2); t4=t4*t4;
   *l2 =sqrt(t1+t2+t4);
   double t5 = (this->A-y-this->m); t5=t5*t5;
   double t6 = (this->A/2 - x); t6 *= t6;
   *l3 =sqrt(t1+t5+t6);
   double t7 = (y-this->A/2); t7 *= t7;
   *l4 =sqrt((z-this->d)*(z-this->d) + t7 + t6);
}

void Seilbot::setZeroPosition(double x0,double y0,double z0)
{
  this->x0=x0;
  this->y0=y0;
  this->z0=z0;
  this->calcLengths(x0,y0,z0,&this->l1_0,&this->l2_0,&this->l3_0,&this->l4_0);    
}

int Seilbot::lenToCounts(double dl)
{
   return (int) (0.5+ 	dl / this->lpr * this->cpr * this->factor);
   //return (int) (0.5+  dl /0.19); //Kette
}

void Seilbot::calcDeltaLengths(double x, double y, double z, double * l1,double * l2, double *l3, double * l4)
{
   double cl1,cl2,cl3,cl4;
   this->calcLengths(x,y,z,&cl1,&cl2,&cl3,&cl4);
  *l1= cl1-this->l1_0;
  *l2= cl2-this->l2_0;
  *l3= cl3-this->l3_0;
  *l4= cl4-this->l4_0;
}
 
void Seilbot::calcDeltaCounts(double x, double y, double z, int *c1, int * c2, int *c3, int*c4)
{
   double dl1,dl2,dl3,dl4;
   this->calcDeltaLengths(x,y,z,&dl1,&dl2,&dl3,&dl4);
   *c1= this->lenToCounts(dl1);
   *c2= this->lenToCounts(dl2);
   *c3= this->lenToCounts(dl3);
   *c4= this->lenToCounts(dl4);
}

#ifdef MAIN
int main(int argc,char ** argv)
{  lpr= (r + delta/2)*M_PI;
   double x,y,z;
   double l1,l2,l3,l4;
   double l1_0,l2_0,l3_0,l4_0;
   calcLengths(A/2,A/2,0,&l1_0, &l2_0, &l3_0, &l4_0);


   x=atof(argv[1]);
   y=atof(argv[2]);
   z=atof(argv[3]);
   printf("%lf %lf %lf\n",x,y,z);
   calcLengths(x,y,z,&l1, &l2,&l3,&l4);
   printf("%lf %lf %lf %lf\n",l1,l2,l3,l4);

   printf("%lf %lf %lf %lf\n",l1-l1_0,l2-l2_0,l3-l3_0,l4-l4_0);
   printf("%d %d %d %d\n",calcCounts(l1-l1_0),calcCounts(l2-l2_0), calcCounts(l3-l3_0),calcCounts(l4-l4_0));
   return 0;
}
#endif
